Require Import Bool.
Require Import String.
Require Import Classical.

Open Scope string_scope.


(** Definition of classical realizability, compile with Coq v. trunk (14850) for kextraction (is it useful?) **)

(* To solve bug 2630. *)
Ltac easy ::=
  let rec use_hyp H :=
   (match type of H with
    | _ /\ _ => exact H || destruct_hyp H
    | _ => try (solve [ inversion H ])
    end)
  with do_intro := (let H := fresh in
                    intro H; use_hyp H)
  with destruct_hyp H := (case H; clear H; do_intro; do_intro) in
  let rec use_hyps :=
   (match goal with
    | H:_ /\ _ |- _ => exact H || (destruct_hyp H; use_hyps)
    | H:_ |- _ => solve [ inversion H ]
    | _ => idtac
    end) in
  let rec do_atom :=
   ((solve [ reflexivity | symmetry; trivial ]) || (* the only modified line *)
      contradiction || (split; do_atom))
  with do_ccl := (trivial with eq_true; do_atom || (repeat do_intro; do_atom)) in
  (use_hyps; do_ccl) || fail "Cannot solve this goal".


Ltac exist2 t₁ t₂ := exists t₁; exists t₂. (* cannot use exists2 as it is already used… *)
Ltac exists3 t₁ t₂ t₃ := exists t₁; exists t₂; exists t₃.
Ltac exists4 t₁ t₂ t₃ t₄ := exists t₁; exists t₂; exists t₃; exists t₄.
Ltac exists5 t₁ t₂ t₃ t₄ t₅ := exists t₁; exists t₂; exists t₃; exists t₄; exists t₅.

(*
Instance not_proper {T} (R : relation T) `{Symmetric _ R} : Symmetric (fun x y => ~R x y).
Proof. intros x y Hxy HR. now elim Hxy. Qed.
Class Empty_set (T : Type) := {emptyset : T}.
Notation "'∅'" := emptyset.
*)

(* When the Timing extension is not present. *)
Ltac start_timer c := idtac.
Ltac stop_timer c := idtac.


(*************************)
(**  Krivine's machine  **)
(*************************)


(** An abstract set of instructions, containing callcc **)
Parameter instruction : Set.
Parameter callcc : instruction.

(** A set of stack constants **)
Parameter stack_const : Set.

(** Λ_c terms **)
Inductive term : Set :=
  | Cst : instruction -> term
  | Lam : string -> term -> term
  | Var : string -> term
  | App : term -> term -> term.
Coercion Cst : instruction >-> term.
Coercion Var : string >-> term.
Notation "'λ' n t" := (Lam n t) (at level 16, n at level 0, t at level 41, format "'λ' n '/  '  t").
Notation "t @ s" := (App t s) (at level 15, left associativity).


Parameter KAM_const : Set.
(** Closures, stacks and environments **)
Inductive Λ : Set :=
  | Closed : term -> env -> Λ
  | Cont : Π -> Λ
with Π : Set :=
  | Scst : stack_const -> Π
  | Scons : Λ -> Π -> Π
with env : Set :=
  | Enil : env
  | Econs : string -> Λ -> env -> env.
Coercion Scst : stack_const >-> Π.
Notation "t ↓ e" := (Closed t e) (at level 45, no associativity, format "t '/  ' '↓' e").
Notation "k[ π ]" := (Cont π).
Notation "t · s" := (Scons t s) (at level 47, right associativity).
Notation "e ← t ; f" := (Econs e t f) (at level 49, right associativity, format "e '←' t ';'  '/' f").
Notation "∅" := Enil.

Bind Scope Env_scope with env.
Bind Scope Stack_scope with Π.
Bind Scope Clos_scope with Λ.

Arguments Scope Cont [Stack_scope].
Arguments Scope Scons [Clos_scope Stack_scope].
Arguments Scope Econs [Clos_scope Env_scope].

(* Extracting a closure from an environment *)
Fixpoint get s (e : env) : Λ :=
  match e with
    | c←t;e' => if string_dec s c then t else get s e'
    | Enil => "dummy"↓Enil (* dummy case *)
  end.

(* To have a general reduction rule for Application compatible with native data
Definition close t e :=
  match t with
    | Var n => get n e
    | _ => t↓e
  end.*)

Fixpoint closed_aux acc t :=
  match t with
    | Cst _ => true
    | Lam s t' => closed_aux (cons s acc) t'
    | Var n => if List.in_dec string_dec n acc then true else false
    | App t₁ t₂ => closed_aux acc t₁ && closed_aux acc t₂
  end.

Definition closed := closed_aux nil.


(** Processes **)
Inductive process := Process : Λ -> Π -> process.
Notation "c '★' s" := (Process c s) (at level 55).

(** **  Reduction rules  **)

Parameter red : process -> process -> Prop.
Axiom red_trans : forall p₁ p₂ p₃, red p₁ p₂ -> red p₂ p₃ -> red p₁ p₃.
(* reduction rules: grab, push, save, restore plus one for variables and environments *)
Axiom red_Lam : forall n t c e π, red ((λ n t)↓e ★ c·π) (t↓(n←c;e) ★ π).
Axiom red_App : forall t t' e π, red ((t @ t')↓e ★ π) (t↓e ★ t'↓e · π).
Axiom red_cc : forall t e π, red (callcc↓e ★ t·π) (t ★ k[π]·π).
Axiom red_k : forall t π π', red (k[π] ★ t·π') (t ★ π).
Axiom red_Var : forall n e π, red (Var n↓e ★ π) (get n e ★ π).
(* This last rule is necessary for native data: we unbox data to expose it to the operators *)
Axiom red_AppVar : forall t n e π, red ((t @ Var n)↓e ★ π) (t↓e ★ get n e·π).

(* Printing command for λ_c terms. *)
Section Printing.
  Variable print_const : instruction -> string.
  Variable print_sconst : stack_const -> string.
  Variable print_KAM_const : KAM_const -> string.

  Fixpoint print_term (t : term) : string :=
    match t with
      | Cst c => print_const c
      | Lam n t => "λ" ++ n ++ " " ++ print_term t
      | Var n => n
      | App t₁ t₂ => print_term t₁ ++ " " ++ print_term t₂
    end.
  
  Fixpoint print_clos c :=
    match c with
      | Closed t e => (print_term t ++ "↓" ++ print_env e)%string
      | Cont s => "k[" ++ print_stack s ++ "]"
    end
  with print_stack s :=
    match s with
      | Scst α => print_sconst α
      | Scons c s' => print_clos c ++ "°" ++ print_stack s'
    end
  with print_env e :=
    match e with
      | Enil => "∅"%string
      | Econs n c Enil => n ++ " ↦ " ++ print_clos c
      | Econs n c e' => n ++ " ↦ " ++ print_clos c ++ ", " ++ print_env e'
    end.
End Printing.

(** **  Some usual terms  **)

Definition Id := λ"x" "x".
Definition nId := λ"x" ("x" @ "x"). (* or any λx. x u *)
Definition tt := λ"x" λ"y" "x".
Definition ff := λ"x" λ"y" "y".
(** A universal realizer of excluded middle **)
Definition em := λ"f" λ"g" callcc @ λ"k" "g" @ λ"i" "k" @ ("f" @ "i").
(** Turing's fixpoint operator **)
Definition Y := (λ"x" λ"y" "y" @ ("x" @ "x" @ "y")) @ (λ"x" λ"y" "y" @ ("x" @ "x" @ "y")).

(** Storage operators for functions.
    We impose the presence of the continuation(s) k (or u and v)
    to ensure that reduction can only occur when it is present. **)
Definition caron1 op := λ"Mx" "Mx" @ λ"x" op @ "x".
Definition caron2 op := λ"Mx" "Mx" @ λ"x" λ"My" "My" @ λ"y" op @ "x" @ "y".
Definition caron3 op := λ"Mx" "Mx" @ λ"x" λ"My" "My" @ λ"y" λ"Mz" "Mz" @ λ"z" op @ "x" @ "y" @ "z".
Definition caron4 op := λ"Mx" "Mx" @ λ"x" λ"My" "My" @ λ"y" λ"Mz" "Mz" @ λ"z"
  λ"Mt" "Mt" @ λ"t" op @ "x" @ "y" @ "z" @ "t".

(** Same thing for relations. **)
Definition rel_caron1 rel := λ"Mx" "Mx" @ λ"x" rel @ "x".
Definition rel_caron2 rel := λ"Mx" "Mx" @ λ"x" λ"My" "My" @ λ"y" rel @ "x" @ "y".
Definition rel_caron3 rel :=
  λ"Mx" "Mx" @ λ"x" λ"My" "My" @ λ"y" λ"Mz" "Mz" @ λ"z" rel @ "x" @ "y" @ "z".
Definition rel_caron4 rel := λ"Mx" "Mx" @ λ"x" λ"My" "My" @ λ"y" λ"Mz" "Mz" @ λ"z" λ"Mt" "Mt" @ λ"t"
  rel @ "x" @ "y" @ "z" @ "t".


(**************************)
(**  Second order Logic  **)
(**************************)


(* Through a shallow embedding:
   defined by falsity values, thus after realizability to bypass the adequacy lemma *)


(***********************)
(** *  Realizability  **)
(***********************)


Parameter pole : process -> Prop.
Notation "p ∈ ⫫" := (pole p) (at level 69, format "p  '∈'  '⫫'").
Axiom anti_evaluation : forall p p', red p p' -> p' ∈ ⫫ -> p ∈ ⫫.

Definition formula := Π -> Prop.

Definition Fval (F : formula) (π : Π) := F π.
Notation "π '∈' '‖' F '‖'" := (Fval F π) (at level 70, format "'[hv' π  '∈' '/'  '‖' F '‖' ']'").

Definition realizes t F := forall π, π ∈ ‖F‖ -> t★π ∈ ⫫.
Notation "t ⊩ F" := (realizes t F) (at level 79).

Definition Impl A B := fun π => match π with Scst _ => False | t·π' => t ⊩ A /\ π' ∈ ‖B‖ end.
Notation "A '→' B" := (Impl A B) (at level 78, right associativity).


Ltac clear_realizers :=
  repeat lazymatch goal with
    | H : _ ⊩ _ |- _ => clear H
    | H : _ ∈ ‖_‖ |- _ => clear H
    | π : Π |- _ => clear π
    | t : Λ |- _ => clear t
    | e : env |- _ => clear e
  end.

(** **  Quantifications  **)

(** *** Without relativisation **)
Definition Forall T f : formula := fun π => exists t : T, π ∈ ‖f t‖.
Global Notation "'∀' t₁ .. t₂ ',' F" := (Forall _ (fun t₁ => .. (Forall _ (fun t₂ => F)) .. ))
  (at level 99, t₁ binder, t₂ binder, right associativity).
Global Notation "'∃' t₁ .. t₂ ',' F" :=
  (Forall formula (fun Z => (Forall _ (fun t₁ => .. (Forall _ (fun t₂ => F → Z)) ..)) → Z))
  (at level 99, t₁ binder, t₂ binder, right associativity).

(** *** With relativisation **)
Class Relativisation {T : Type} (P : T -> formula) := Rel : (T -> formula) -> T -> formula.
Global Instance now_Rel {T} P f : @Relativisation T P := {Rel := f}.
Global Instance make_Rel {T} P : @Relativisation T P := {Rel := fun f t => P t → f t}.
Global Hint Unfold Rel now_Rel make_Rel : Krivine.
(* Note: We cannot add the equivalence between relativized formula and the formula with the predicate
         as a precondition because the idea is precisely to modify the number of arguments
         so that terms cannot be equivalent.
         We could add the existence of terms performing the translation in both directions as a way to ensure
         correctness but this could not be used. *)

(** Binding 1 variable **)
Definition ForallR {T} P `{@Relativisation T P} (f : T -> formula) : formula :=
  fun π => exists t : T, π ∈ ‖Rel f t‖.
Global Notation "'∀₁' t '∈' P ',' F" := (ForallR P (fun t => F)) (at level 99, right associativity).
Global Notation "'∃₁' t '∈' P ',' F" :=
  (Forall formula (fun Z => (ForallR P (fun t => F → Z)) → Z)) (at level 99, t at next level).

(** Binding 2 variables **)
Definition ForallR2 {T₁ T₂} P₁ P₂ `{@Relativisation T₁ P₁, @Relativisation T₂ P₂} (f : T₁ -> T₂ -> formula)
  : formula := fun π => exists t₁ : T₁, exists t₂ : T₂, π ∈ ‖Rel (fun t => Rel (f t) t₂) t₁‖.
Global Notation "'∀₂' t₁ ',' t₂ '∈' P₁ '×' P₂ ',' F" :=
  (ForallR2 P₁ P₂ (fun t₁ t₂ => F)) (at level 99, right associativity).
Global Notation "'∃₂' t₁ ',' t₂ '∈' P₁ '×' P₂ ',' F" :=
  (Forall formula (fun Z => (ForallR2 P₁ P₂ (fun t₁ t₂ => F → Z)) → Z)) (at level 99, right associativity).

(** Binding 3 variables **)
Definition ForallR3 {T₁ T₂ T₃} P₁ P₂ P₃ `{@Relativisation T₁ P₁, @Relativisation T₂ P₂, @Relativisation T₃ P₃}
                    (f : T₁ -> T₂ -> T₃ -> formula) : formula :=
  fun π => exists t₁ : T₁, exists t₂ : T₂, exists t₃ : T₃,
           π ∈ ‖Rel (fun t => Rel (fun t' => Rel (f t t') t₃) t₂) t₁‖.
Global Notation "'∀₃' t₁ ',' t₂ ',' t₃ '∈' P₁ '×' P₂ '×' P₃ ',' F" :=
  (ForallR3 P₁ P₂ P₃ (fun t₁ t₂ t₃ => F)) (at level 99,  right associativity).
Global Notation "'∃₃' t₁ ',' t₂ ',' t₃ '∈' P₁ '×' P₂ '×' P₃ ',' F" :=
  (Forall formula (fun Z => (ForallR3 P₁ P₂ P₃ (fun t₁ t₂ t₃ => F → Z)) → Z))
  (at level 99, right associativity).

(** Binding 4 variables **)
Definition ForallR4 {T₁ T₂ T₃ T₄} P₁ P₂ P₃ P₄
                    `{@Relativisation T₁ P₁, @Relativisation T₂ P₂, @Relativisation T₃ P₃, @Relativisation T₄ P₄}
                    (f : T₁ -> T₂ -> T₃ -> T₄ -> formula) : formula :=
  fun π => exists t₁ : T₁, exists t₂ : T₂, exists t₃ : T₃, exists t₄ : T₄,
           π ∈ ‖Rel (fun t => Rel (fun t' => Rel (fun t'' => Rel (f t t' t'') t₄) t₃) t₂) t₁‖.
Global Notation "'∀₄' t₁ ',' t₂ ',' t₃ ',' t₄ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ ',' F" :=
  (ForallR4 P₁ P₂ P₃ P₄ (fun t₁ t₂ t₃ t₄ => F)) (at level 99,  right associativity).
Global Notation "'∃₄' t₁ ',' t₂ ',' t₃ ',' t₄ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ ',' F" :=
  (Forall formula (fun Z => (ForallR4 P₁ P₂ P₃ P₄ (fun t₁ t₂ t₃ t₄ => F → Z)) → Z))
  (at level 99, right associativity).

(** Binding 5 variables **)
Definition ForallR5 {T₁ T₂ T₃ T₄ T₅} P₁ P₂ P₃ P₄ P₅
  `{@Relativisation T₁ P₁, @Relativisation T₂ P₂, @Relativisation T₃ P₃, @Relativisation T₄ P₄,
    @Relativisation T₅ P₅} (f : T₁ -> T₂ -> T₃ -> T₄ -> T₅ -> formula) : formula :=
  fun π => exists t₁ : T₁, exists t₂ : T₂, exists t₃ : T₃, exists t₄ : T₄, exists t₅ : T₅,
    π ∈ ‖Rel (fun t => Rel (fun t' => Rel (fun t'' => Rel (fun t''' => Rel (f t t' t'' t''') t₅) t₄) t₃) t₂) t₁‖.
Global Notation "'∀₅' t₁ ',' t₂ ',' t₃ ',' t₄ ',' t₅ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ '×' P₅ ',' F" :=
  (ForallR5 P₁ P₂ P₃ P₄ P₅ (fun t₁ t₂ t₃ t₄ t₅ => F)) (at level 99,  right associativity).
Global Notation "'∃₅' t₁ ',' t₂ ',' t₃ ',' t₄ ',' t₅ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ '×' P₅ ',' F" :=
  (Forall formula (fun Z => (ForallR5 P₁ P₂ P₃ P₄ P₅ (fun t₁ t₂ t₃ t₄ t₅ => F → Z)) → Z))
  (at level 99, right associativity).

(** Function relativisation **)
Definition fun_Pred {T T'} P P' (f : T -> T') := ∀₁x∈P, P' (f x).
Global Instance fun_Rel {T T'} P P' `{Relativisation T P} : @Relativisation (T -> T') (fun_Pred P P') :=
  now_Rel (fun_Pred P P') (fun f t => (∀₁x∈P, P'(t x)) → f t).
Notation "A '~>' B" := (fun_Pred A B) (at level 80, right associativity).


(** ** Usual connectives **)

Definition Top : formula := fun _ => False.
Notation "⊤" := Top.

Definition bot := ∀ Z, Z.
Notation "⊥" := bot.

Notation "¬ F" := (F → ⊥) (at level 25, only parsing).

Definition one := ∀ Z, (Z → Z).


(** Implicational equality **)
(** Be careful that the equivalence between c → A and c ↦ A can only be realized when c is realized
    by the identity. This means that taking c = n <> m does not provide the equivalence.
    In this case, it is better to take a boolean equality eqb since x <> y then amounts to eqb(x,y) = 0. **)
Definition mapsto (c : Prop) F := fun π => c /\ π ∈ ‖F‖.
Notation "c '↦' F" := (mapsto c F) (at level 78, right associativity).

(** Intersection type **)
Definition inter A B := fun π => π ∈ ‖A‖ \/ π ∈ ‖B‖.
Notation "A '∩' B" := (inter A B) (at level 71, right associativity).


(** Optimized version for n-ary and **)
Definition and2 A B := ∀ Z, (A → B → Z) → Z.
Notation "A ∧ B" := (and2 A B) (at level 80, no associativity).
Definition and3 A B C := ∀ Z, (A → B → C → Z) → Z.
Notation "A ∧ B ∧ C" := (and3 A B C) (at level 80, B at next level, no associativity).
Definition and4 A B C D := ∀ Z, (A → B → C → D → Z) → Z.
Notation "A ∧ B ∧ C ∧ D" := (and4 A B C D) (at level 80, B at next level, C at next level, no associativity).
Definition and5 A B C D E := ∀ Z, (A → B → C → D → E → Z) → Z.
Notation "A ∧ B ∧ C ∧ D ∧ E" := (and5 A B C D E)
  (at level 80, B at next level, C at next level, D at next level, no associativity).
Definition and6 A B C D E F := ∀ Z, (A → B → C → D → E → F → Z) → Z.
Notation "A ∧ B ∧ C ∧ D ∧ E ∧ F" := (and6 A B C D E F)
  (at level 80, B at next level, C at next level, D at next level, E at next level, no associativity).

(** Optimized versions for n-ary or **)
Definition or2 A B := ∀ Z, (A → Z) → (B → Z) → Z.
Notation "A ∨ B" := (or2 A B) (at level 85, no associativity).
Definition or3 A B C := ∀ Z, (A → Z) → (B → Z) → (C → Z) → Z.
Notation "A ∨ B ∨ C" := (or3 A B C) (at level 85, B at next level, no associativity).
Definition or4 A B C D := ∀ Z, (A → Z) → (B → Z) → (C → Z) → (D → Z) → Z.
Notation "A ∨ B ∨ C ∨ D" := (or4 A B C D) (at level 85, B at next level, C at next level, no associativity).
Definition or5 A B C D E := ∀ Z, (A → Z) → (B → Z) → (C → Z) → (D → Z) → (E → Z) → Z.
Notation "A ∨ B ∨ C ∨ D ∨ E" := (or5 A B C D E)
  (at level 85, B at next level, C at next level, D at next level, no associativity).
Definition or6 A B C D E F := ∀ Z, (A → Z) → (B → Z) → (C → Z) → (D → Z) → (E → Z) → (F → Z) → Z.
Notation "A ∨ B ∨ C ∨ D ∨ E ∨ F" := (or6 A B C D E F)
  (at level 85, B at next level, C at next level, D at next level, E at next level, no associativity).
(* 
Notation "A '∨' .. '∨' B" := (Forall formula Z, (A → Z) → .. ((B → Z) → Z)..).
Error: A recursive notation must start with at least one symbol.
*)


(** Combining exitential quantifiers and conjonction **)
(* Conflict with the other ∀_/∃_ notation without { } *)

(* Without ↦ *)
Global Notation "'Ex₁' t '∈' P ',' '{' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR P (fun t => F₁ → .. (F₂ → Z) .. )) → Z))
  (at level 98, t at next level, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₂' t₁ ',' t₂ '∈' P₁ '×' P₂ ',' '{' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR2 P₁ P₂ (fun t₁ t₂ => F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₃' t₁ ',' t₂ ',' t₃ '∈' P₁ '×' P₂ '×' P₃ ',' '{' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR3 P₁ P₂ P₃ (fun t₁ t₂ t₃ => F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₄' t₁ ',' t₂ ',' t₃ ',' t₄ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ ',' '{' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR4 P₁ P₂ P₃ P₄ (fun t₁ t₂ t₃ t₄ => F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F₁ at level 79, F₂ at level 79).
Global Notation
  "'Ex₅' t₁ ',' t₂ ',' t₃ ',' t₄ ',' t₅ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ '×' P₅ ',' '{' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR5 P₁ P₂ P₃ P₄ P₅ (fun t₁ t₂ t₃ t₄ t₅ => F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F₁ at level 79, F₂ at level 79).

(* With one ↦ *)
Global Notation "'Ex₁' t '∈' P ',' '{' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR P (fun t => F₀ ↦ F₁ → .. (F₂ → Z) .. )) → Z))
  (at level 98, t at next level, F₀ at level 79, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₂' t₁ ',' t₂ '∈' P₁ '×' P₂ ',' '{' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR2 P₁ P₂ (fun t₁ t₂ => F₀ ↦ F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F₀ at level 79, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₃' t₁ ',' t₂ ',' t₃ '∈' P₁ '×' P₂ '×' P₃ ',' '{' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR3 P₁ P₂ P₃ (fun t₁ t₂ t₃ => F₀ ↦ F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F₀ at level 79, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₄' t₁ ',' t₂ ',' t₃ ',' t₄ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ ',' '{' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR4 P₁ P₂ P₃ P₄ (fun t₁ t₂ t₃ t₄ => F₀ ↦ F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F₀ at level 79, F₁ at level 79, F₂ at level 79).
Global Notation
  "'Ex₅' t₁ ',' t₂ ',' t₃ ',' t₄ ',' t₅ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ '×' P₅ ',' '{' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR5 P₁ P₂ P₃ P₄ P₅ (fun t₁ t₂ t₃ t₄ t₅ => F₀ ↦ F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F₀ at level 79, F₁ at level 79, F₂ at level 79).

(* With two ↦ *)
Global Notation "'Ex₁' t '∈' P ',' '{' F '&' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR P (fun t => F ↦ F₀ ↦ F₁ → .. (F₂ → Z) .. )) → Z))
  (at level 98, t at next level, F at level 79, F₀ at level 79, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₂' t₁ ',' t₂ '∈' P₁ '×' P₂ ',' '{' F '&' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR2 P₁ P₂ (fun t₁ t₂ => F ↦ F₀ ↦ F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F at level 79, F₀ at level 79, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₃' t₁ ',' t₂ ',' t₃ '∈' P₁ '×' P₂ '×' P₃ ',' '{' F '&' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR3 P₁ P₂ P₃ (fun t₁ t₂ t₃ => F ↦ F₀ ↦ F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F at level 79, F₀ at level 79, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₄' t₁ ',' t₂ ',' t₃ ',' t₄ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ ',' '{' F '&' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR4 P₁ P₂ P₃ P₄ (fun t₁ t₂ t₃ t₄ => F ↦ F₀ ↦ F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F at level 79, F₀ at level 79, F₁ at level 79, F₂ at level 79).
Global Notation
  "'Ex₅' t₁ ',' t₂ ',' t₃ ',' t₄ ',' t₅ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ '×' P₅ ',' '{' F '&' F₀ '&' F₁ '∧' .. '∧' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR5 P₁ P₂ P₃ P₄ P₅ (fun t₁ t₂ t₃ t₄ t₅ => F ↦ F₀ ↦ F₁ → .. (F₂ → Z) ..)) → Z))
  (at level 98, right associativity, F at level 79, F₀ at level 79, F₁ at level 79, F₂ at level 79).

(** Combining existential quantifiers and disjonction **)
Global Notation "'Ex₁' t '∈' P ',' '{' F₁ '∨' .. '∨' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR P (fun t => (F₁ → Z) → .. ((F₂ → Z) → Z) .. )) → Z))
  (at level 98, t at next level, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₂' t₁ ',' t₂ '∈' P₁ '×' P₂ ',' '{' F₁ '∨' .. '∨' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR2 P₁ P₂ (fun t₁ t₂ => (F₁ → Z) → .. ((F₂ → Z) → Z) .. )) → Z))
  (at level 98, right associativity, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₃' t₁ ',' t₂ ',' t₃ '∈' P₁ '×' P₂ '×' P₃ ',' '{' F₁ '∨' .. '∨' F₂ '}'" :=
  (Forall formula (fun Z => (ForallR3 P₁ P₂ P₃ (fun t₁ t₂ t₃ => (F₁ → Z) → .. ((F₂ → Z) → Z) .. )) → Z))
  (at level 98, right associativity, F₁ at level 79, F₂ at level 79).
Global Notation "'Ex₄' t₁ ',' t₂ ',' t₃ ',' t₄ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ ',' '{' F₁ '∨' .. '∨' F₂ '}'" :=
  (Forall formula
    (fun Z => (ForallR4 P₁ P₂ P₃ P₄ (fun t₁ t₂ t₃ t₄ => (F₁ → Z) → .. ((F₂ → Z) → Z) .. )) → Z))
  (at level 98, right associativity, F₁ at level 79, F₂ at level 79).
Global Notation
  "'Ex₅' t₁ ',' t₂ ',' t₃ ',' t₄ ',' t₅ '∈' P₁ '×' P₂ '×' P₃ '×' P₄ '×' P₅ ',' '{' F₁ '∨' .. '∨' F₂ '}'" :=
  (Forall formula
    (fun Z => (ForallR5 P₁ P₂ P₃ P₄ P₅ (fun t₁ t₂ t₃ t₄ t₅ => (F₁ → Z) → ..((F₂ → Z) → Z)..)) → Z))
  (at level 98, right associativity, F₁ at level 79, F₂ at level 79).


(********************************)
(** *  Automatisation tactics  **)
(********************************)

(** The tactics meant to be used are:
    - [ok]/[eok] for solving very simple goals
    - [Keval] for performing a one-step reduction by the KAM
    - [Kevals] to evaluate as much as possible
    - [dstack] to exhibit the first element of a stack when there is one
    - [dec] to exhibit all elements of a stack
    - [start] to start a proof, basically combines dec and Kevals
    - [find] to guess witnesses by meta-variables and try to solve the immediate goals
    - [Ksolve] automatic procedure that can solve simple proof
    - [Kmove] same as [Ksolve] but does not make guesses with [find]

    The tactics [start] and [find] have variants [startn] and [findn] where you indicate the number of steps/guesses.
    This avoids too eager unfolding and bad guesses.

    When introducing new connectives, you need to extend (through redefinition) the tactics [dstack] and
    [Keval] for dealing with them.  They are written in CPS style so that you can choose the order of basic tactics
    to improve performance.
    All the other tactics will take advantage of the modifications.
**)

(** A closing tactic which proves obvious statements:
    - [t ⊩ A,  π ∈ ‖A‖  |-  t ★ π ∈ ⫫]
    - [_  |-  π ∈ ‖⊥‖]
    - [_  |-  t ⊩ ⊤]
    - [π ∈ ‖⊤‖ |- _]
    - anything solved by [tac]
**)
Lemma Var_subst : forall n e t F, t ⊩ F -> get n e = t -> n↓e ⊩ F.
Proof. intros n e t F Ht Hget π Hπ. eapply anti_evaluation. apply red_Var. rewrite Hget. now apply Ht. Qed.

Ltac end_tac tac :=
  lazymatch goal with
    | [H : ?t ⊩ ?A |- Closed (Var _) _ ⊩ ?A] => apply Var_subst with t; reflexivity || now apply H
    | [H₁ : ?t ⊩ ?A, H₂ : ?π ∈ ‖?A‖ |- ?t ★ ?π ∈ ⫫] => now apply H₁
    | [ |- ?π ∈ ‖⊥‖] => now exists (fun _ : Π => True)
    | [ |- ?t ⊩ ⊤] => intros ? ?; contradiction
    | [H : ?π ∈ ‖⊤‖ |- _] => elim H
    | _ => tac
  end.

Ltac ok := end_tac assumption.

(** The same as the previous one but uses eassumption instead of assumption **)
Ltac eok := end_tac eassumption.

(** Automatically fill in existentials and try to realize them with eok **)
Ltac find := repeat (eexists; repeat split; try eok).
Tactic Notation "findn" integer(n) := do n (eexists; repeat split; try eok).

(* For debugging: redefine it by idtac str to print the debug string. *)
Ltac Debug str := idtac.

(* Basic fragment of a tactic for applying anti_evaluation
   We choose not to unfold definitions because they usually have their own realizability lemmata. *)
Ltac basic_Keval tac :=
  start_timer "basic_Keval";
  lazymatch goal with
    | [ |- (Lam ?n ?t)↓ ?e ★ ?t₁·?s ∈ ⫫] =>
         Debug "Keval_Lam"; apply anti_evaluation with (t↓(n←t₁;e) ★ s); [now apply red_Lam |]
    | [ |- (App ?t₁ (Var ?n))↓ ?e ★ ?s ∈ ⫫] =>
         Debug "Keval_AppVar"; apply anti_evaluation with (t₁↓e ★ (get n e · s)); [now apply red_AppVar | simpl get]
    | [ |- (App ?t₁ ?t₂)↓ ?e ★ ?s ∈ ⫫] =>
         Debug "Keval_Var"; apply anti_evaluation with (t₁↓e ★ (t₂↓e · s)); [now apply red_App |]
    | [ |- Cst callcc↓ ?e ★ ?t·?s ∈ ⫫] =>
         Debug "Keval_callcc"; apply anti_evaluation with (t ★ k[s] · s); [now apply red_cc |]
    | [ |- (Cont ?s) ★ ?t·_ ∈ ⫫] =>
         Debug "Keval_k"; apply anti_evaluation with (t ★ s); [now apply red_k |]
    | [ |- (Var ?n)↓ ?e ★ ?s ∈ ⫫] =>
         Debug "Keval_Var"; apply anti_evaluation with (get n e ★ s); [now apply red_Var | simpl get]
    | _ => tac
  end;
  stop_timer "basic_Keval".

(* Basic fragment of a tactic for destructing stack depending on its falsity value *)
Ltac basic_dstack tac Hπ :=
  start_timer "basic_dstack";
  lazymatch type of Hπ with
    | context[Rel _ _] => repeat autounfold with Krivine in Hπ
    (*repeat (unfold Rel in Hπ  || unfold now_Rel in Hπ || unfold make_Rel in Hπ)*)
    | ?π ∈ ‖ _ → _ ‖ =>
      let t := fresh "t" in let Ht := fresh "Ht" in
      destruct π as [| t π]; [contradiction | simpl in Hπ; destruct Hπ as [Ht Hπ]]
    | ?π ∈ ‖ ∀ t, _ ‖ => let tt := fresh t in destruct Hπ as [tt Hπ]
    | ?π ∈ ‖ ∃ _, _ ‖ => let Z := fresh "Z" in destruct Hπ as [Z Hπ]
    | ?π ∈ ‖ ∀₁ t ∈ _, _ ‖ => let tt := fresh t in destruct Hπ as [tt Hπ]
    | ?π ∈ ‖ ∃₁ _ ∈ _, _ ‖ => let Z := fresh "Z" in destruct Hπ as [Z Hπ]
    | ?π ∈ ‖ ∀₂ t₁, t₂ ∈ _ × _, _ ‖ => let tt₁ := fresh t₁ in let tt₂ := fresh t₂ in destruct Hπ as [tt₁ [tt₂ Hπ]]
    | ?π ∈ ‖ ∃₂ _, _ ∈ _ × _, _ ‖ => let Z := fresh "Z" in destruct Hπ as [Z Hπ]
    | ?π ∈ ‖ ∀₃ t₁, t₂, t₃ ∈ _ × _ × _, _ ‖ =>
      let tt₁ := fresh t₁ in let tt₂ := fresh t₂ in let tt₃ := fresh t₃ in destruct Hπ as [tt₁ [tt₂ [tt₃ Hπ]]]
    | ?π ∈ ‖ ∃₃ _, _, _ ∈ _ × _ × _, _ ‖ => let Z := fresh "Z" in destruct Hπ as [Z Hπ]
    | ?π ∈ ‖ ∀₄ t₁, t₂, t₃, t₄ ∈ _ × _ × _ × _, _ ‖ =>
      let tt₁ := fresh t₁ in let tt₂ := fresh t₂ in let tt₃ := fresh t₃ in let tt₄ := fresh t₄ in
      destruct Hπ as [tt₁ [tt₂ [tt₃ [tt₄ Hπ]]]]
    | ?π ∈ ‖ ∃₄ _, _, _, _ ∈ _ × _ × _ × _, _ ‖ => let Z := fresh "Z" in destruct Hπ as [Z Hπ]
    | ?π ∈ ‖ ∀₅ t₁, t₂, t₃, t₄, t₅ ∈ _ × _ × _ × _ × _, _ ‖ =>
      let tt₁ := fresh t₁ in let tt₂ := fresh t₂ in let tt₃ := fresh t₃ in let tt₄ := fresh t₄ in
      let tt₅ := fresh t₅ in destruct Hπ as [tt₁ [tt₂ [tt₃ [tt₄ [tt₅ Hπ]]]]]
    | ?π ∈ ‖ ∃₅ _, _, _, _, _ ∈ _ × _ × _ × _ × _, _ ‖ => let Z := fresh "Z" in destruct Hπ as [Z Hπ]
    | ?π ∈ ‖ and2 _ _ ‖ => unfold and2 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ and3 _ _ _ ‖ => unfold and3 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ and4 _ _ _ _ ‖ => unfold and4 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ and5 _ _ _ _ _ ‖ => unfold and5 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ and6 _ _ _ _ _ ‖ => unfold and6 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ or2 _ _ ‖ => unfold or2 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ or3 _ _ _ ‖ => unfold or3 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ or4 _ _ _ _ ‖ => unfold or4 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ or5 _ _ _ _ _ ‖ => unfold or5 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ or6 _ _ _ _ _ _ ‖ => unfold or6 in Hπ; basic_dstack tac Hπ
    | ?π ∈ ‖ ?c ↦ ?F ‖ => let Hc := fresh "Hc" in destruct Hπ as [Hc Hπ]
    | ?π ∈ ‖ ?A ∩ ?B ‖ => unfold inter in Hπ; destruct Hπ as [Hπ | Hπ]
    | _ => tac Hπ
  end;
  stop_timer "basic_dstack".

(* Effective tactics using the above fragments.
   Are meant to be redefined when extending the reduction rules / realizability connectives *)
Ltac Keval := basic_Keval fail.
Ltac dstack := basic_dstack fail.
(* User decomposition tactics. Will work with new definition of Keval and dstack *)
Ltac Kevals := start_timer "Kevals"; progress (repeat Keval); stop_timer "Kevals".
Ltac dec π :=
  start_timer "dec";
  match goal with | [Hπ : π ∈ ‖ _ ‖ |- _] => repeat dstack Hπ end;
  stop_timer "dec".

Ltac remove_stacks := (* removing useless stacks from context *)
  try match goal with
    | Hπ : ?π ∈ ‖_‖ |- _ =>
       match goal with
         | |- context[π] => fail 1 (* can appear outside a k[_·s], like for instance π ∈ ‖_‖ *)
         | |- _ => try clear dependent π; remove_stacks
       end
    | π : Π |- _ => (clear dependent π; remove_stacks) || fail 1
    | _ => idtac
  end.

Ltac unfold_def :=
  match goal with
    | |- (Lam _ _)↓ _ ★ _ ∈ ⫫ => idtac
    | |- (App _ _)↓ _ ★ _ ∈ ⫫ => idtac
    | |- Cst callcc↓ _ ★ _ ∈ ⫫ => idtac
    | |- (Cont _) ★ _ ∈ ⫫ => idtac
    | |- ?t↓ _ ★ _ ∈ ⫫ => unfold t at 1; unfold_def
    | |- ?t _↓ _ ★ _ ∈ ⫫ => unfold t at 1; unfold_def
    | |- ?t _ _↓ _ ★ _ ∈ ⫫ => unfold t at 1; unfold_def
    | |- ?t _ _ _↓ _ ★ _ ∈ ⫫ => unfold t at 1; unfold_def
    | |- ?t _ _ _ _↓ _ ★ _ ∈ ⫫ => unfold t at 1; unfold_def
    | |- _ => idtac
  end.

Ltac start :=
  start_timer "start";
  remove_stacks; let π := fresh "π" in let Hπ := fresh "Hπ" in intros π Hπ;
  dec π; unfold_def; (*red_simpl; for close/get*) try Kevals;
  stop_timer "start".

Tactic Notation "startn" integer(n) :=
  remove_stacks; let π := fresh "π" in let Hπ := fresh "Hπ" in intros π Hπ;
  dec π; unfold_def; (*red_simpl;*) do n Keval.

Ltac Kprogress := (* TODO: optimize the case of H : ?s ∈ ‖?A‖ |- _ ∈ ‖∃ _‖ with exists A; split; [| exact H] *)
  match goal with
    | Ht : ?t ⊩ ?A |- ?t ★ _ ∈ ⫫ => apply Ht
    | Hπ : ?s ∈ ‖_‖ |- _ => progress dec s
    | _ => ok || Kevals || (repeat split)
  end.
Ltac Kmove := start_timer "Kmove"; intros; try start; repeat Kprogress; stop_timer "Kmove".

Ltac Ksolve := start_timer "Ksolve"; Kmove; find; repeat Kprogress; stop_timer "Ksolve".
(*
Ltac Ksolve :=
  intros; try start; repeat match goal with
    | Hπ : ?s ∈ ‖_‖ |- context[?s] => progress dec s (* same comment as for remove_stacks *)
    | Ht : ?t ⊩ ?A |- ?t ★ _ ∈ ⫫ => apply Ht
    | _ => ok || Kevals || find
  end.
*)


(*************************************)
(** *  First realizability results  **)
(*************************************)

Lemma Fvalue_Bot : forall π, π ∈ ‖⊥‖.
Proof. intro. ok. Qed.

Lemma Tvalue_Top : forall t, t ⊩ ⊤.
Proof. intro. ok. Qed.

Lemma Id_realizer_by_hand : forall e, Id↓e ⊩ one.
Proof.
unfold one, Id.
intros e π Hπ.
destruct Hπ as [Z Hπ].
destruct π as [|t π'].
  contradiction Hπ.
  destruct Hπ as [Ht Hπ'].
  apply anti_evaluation with (t★π').
    apply red_trans with ("x"↓("x"←t;e) ★ π').
      apply red_Lam.
      apply red_Var.
    apply Ht.
    assumption.
Qed.

Lemma Id_realizer_intermediate : forall e, Id↓e ⊩ one.
Proof.
unfold one, Id.
intros e π Hπ.
dec π.
Kevals.
apply Ht.
assumption.
Qed.

Theorem Id_realizer : forall e, Id↓e ⊩ one.
Proof. unfold one. Ksolve. Qed.

Theorem modus_ponens : forall A B, forall t t' e, t↓e ⊩ A → B -> t'↓e ⊩ A -> t @ t'↓e ⊩ B.
Proof. Ksolve. Qed.

Theorem Impl_eta_realizer :
  forall A B t e, (forall t' e, t'↓e ⊩ A -> t @ t'↓e ⊩ B) -> (λ"x" t @ "x")↓e ⊩ A → B.
Proof. intros A B t e Ht. startn 1. apply (Ht "x"); Ksolve. Qed.

Theorem uniform_Forall : forall (T : Type) (F : T -> formula) t, (t ⊩ ∀n, F n) <-> forall n, t ⊩ F n.
Proof. intros F t. split; intro Ht; Ksolve. now apply (Ht n). Qed.

Lemma k_realizer : forall A B, forall π, π ∈ ‖A‖ -> k[π] ⊩ A → B.
Proof. Ksolve. Qed.

Lemma ex_falso : forall t, forall A, t ⊩ ⊥ -> t ⊩ A.
Proof. Ksolve. Qed.

Theorem callcc_realizes_Peirce : forall e, callcc↓e ⊩ ∀A B, ((A → B) → A) → A.
Proof. Ksolve. now apply k_realizer. Qed.

Theorem excluded_middle : forall e, em↓e ⊩ ∀F, F ∨ ¬F.
Proof. repeat Ksolve. Qed.

Lemma mapsto_equiv : forall c A t, (t ⊩ c ↦ A) <-> (c -> t ⊩ A).
Proof. intros c A t. split; intro Ht. Ksolve. start. now apply Ht. Qed.


(** Turing's fixpoint **)
Theorem Y_realizer T (R : T -> T -> Prop) (Hwf : well_founded R) :
  forall e, Y↓e ⊩ ∀P, (∀x, (∀y, R y x ↦ P y) → P x) → ∀z, P z.
Proof.
intro e. start. revert z π Hπ. induction z as [z Hrec] using (well_founded_ind Hwf); intros.
apply Ht. find. start. eapply Hrec; eok.
Qed.
Arguments Y_realizer [T R] Hwf e π _.

(** Relativized version of Y **)
Definition Y_rel := λ"f" Y @ (λ"x" λ"y" "f" @ "y" @ "x").

Theorem Y_rel_realizer (T : Type) (R : T -> T -> Prop) (Hwf : well_founded R) (rel : T -> formula) :
  forall e, Y_rel↓e ⊩ ∀P, (∀x, rel x → (∀y, rel y → R y x ↦ P y) → P x) → ∀z, rel z → P z.
Proof. intro. startn 2. apply (Y_realizer Hwf). exists (fun t => rel t → P t). repeat Ksolve. Qed.
Arguments Y_rel_realizer [T R] Hwf rel e π _.


(****************************)
(** *  Subtyping relation  **)
(****************************)


Require Import RelationClasses.
Require Import Morphisms.
Set Implicit Arguments.

Definition subtype F F' := forall π, π ∈ ‖F'‖ -> π ∈ ‖F‖.
Notation  "A '⊆' B" := (subtype A B) (at level 80, no associativity).
Transparent subtype.

Definition eqtype F F' := forall π, π ∈ ‖F‖ <-> π ∈ ‖F'‖.
Notation "A '≈' B" := (eqtype A B) (at level 80, no associativity).

(** *** Instances for rewriting **)

(** [A ≈ B] is the equivalence associated with the preorder [A ⊆ B] **)
Instance Eqtype_extensional : Proper (eqtype ==> eq ==> iff) Fval.
Proof. intros A B Heq π π' Heqπ. split; intros Hπ; subst; now rewrite (Heq π') || rewrite <- (Heq π'). Qed.

Instance Eqtype_extensional2 : Proper (eq ==> eqtype ==> iff) realizes.
Proof.
intros t t' Heqt A B Heq. split; intros Ht π Hπ; subst; apply Ht; now apply (Eqtype_extensional Heq eq_refl).
Qed.

Instance Subtype_PreOrder : PreOrder subtype.
Proof. repeat split. now intros ? ? ?. intros A B C HAB HBC π Hπ. apply HAB. now apply HBC. Qed.

Instance Eqtype_equiv : Equivalence eqtype.
Proof.
split.
  intros F s. reflexivity.
  intros A B Heq s. now rewrite Heq.
  intros A B C HAB HBC s. rewrite HAB. now rewrite HBC.
Qed.

Instance Subtype_PO : PartialOrder eqtype subtype.
Proof.
intros x y; split; intro Heq.
  split; intros π Hπ. now rewrite Heq. now rewrite <- Heq.
  destruct Heq as [H₁ H₂]. split; now apply H₁ || apply H₂.
Qed.

Lemma eqtype_subtype : forall A B, A ≈ B -> A ⊆ B.
Proof. intros. rewrite H. reflexivity. Qed.

Lemma subtype_eqtype : forall A B, A ⊆ B -> B ⊆ A -> A ≈ B.
Proof. intros A B HAB HBA s. split; intro Hπ; now (apply HAB || apply HBA). Qed.

(*
(** The connectives respects subtyping. **)
Instance Forall_proper (T : Type) : Proper (pointwise_relation T subtype ==> subtype) (Forall T).
Proof. intros A B HAB π [t Hπ]. exists t. now apply (HAB t). Qed.

Instance mapsto_Proper : Proper (Impl ==> subtype ==> subtype) mapsto.
Proof. intros c₁ c₂ Hcs A B HAB π [Hc Hπ]. split. now rewrite Hcs. now apply HAB. Qed.

Lemma eqtype_subtype : forall A B, A ≈ B -> A ⊆ B.
Proof. intros. rewrite H. reflexivity. Qed.
*)

(** *** Usual properties of a subtyping relation **)

Lemma sub_bot : forall F, ⊥ ⊆ F.
Proof. Ksolve. Qed.

Lemma top_bot_sub_one : (⊤ → ⊥) ⊆ one.
Proof. unfold one. Ksolve. Qed.


Lemma sub_Impl : forall A A' B B', A ⊆ A' -> B ⊆ B' -> A' → B ⊆ A → B'.
Proof. intros A A' B B' HA HB. start. split. start. apply Ht. now apply HA. now apply HB. Qed.


Lemma sub_Forall : forall T A (t : T), (∀x, A x) ⊆ A t.
Proof. Ksolve. Qed.

Lemma Forall_sub_proper : forall T A B, (forall x : T, A x ⊆ B x) -> (∀x, A x) ⊆ ∀x, B x.
Proof. intros T A B Hsub π [n Hn]. exists n. now apply Hsub. Qed.

Lemma sub_Forall_add : forall T A (B : T -> formula), (forall t, A ⊆ B t) -> A ⊆ ∀t, B t.
Proof. intros T A B HAB π [t Hπ]. now apply (HAB t). Qed.

Lemma sub_Forall_cross : forall T A (B : T -> formula), (∀t, A → B t) ≈ A → ∀t, B t.
Proof. intros; split; Ksolve. Qed.

Lemma Forall_comm : forall T₁ T₂ (A : T₁ -> T₂ -> formula), (∀t, ∀u, A t u) ≈ ∀u, ∀t, A t u.
Proof. intros ? ? ? π; split; intro; dec π; find. Qed.

Lemma Forall_cancel : forall T A, inhabited T -> (∀t : T, A) ≈ A.
Proof. intros T A [] π; split; intro; dec π; find; ok. Qed.

Lemma sub_mapsto : forall A c, A ⊆ c ↦ A.
Proof. now intros ? ? ? []. Qed.

Lemma mapsto_sub_proper : forall (c c' : Prop) A B, (c' -> c) -> A ⊆ B -> c ↦ A ⊆ c' ↦ B.
Proof. intros c c' A B Hcs HAB π [Hc Hπ]. split. now apply Hcs. now apply HAB. Qed. 

Lemma mapsto_cancel : forall c A, (c ↦ A) ∩ (~c ↦ A) ≈ A.
Proof.
intros c A π. split; intro Hπ.
  now destruct Hπ as [[]|[]].
  elim (classic c); intro Hc; (now left; split) || (now right; split).
Qed.

Lemma mapsto_idempotent : forall c A, c ↦ c ↦ A ≈ c ↦ A.
Proof. intros. apply subtype_eqtype. now intros ? []; repeat split. apply sub_mapsto. Qed.

Instance mapsto_subtype_proper : Proper (iff ==> eqtype ==> eqtype) mapsto.
Proof.
intros c c' Hc A B HAB π.
split; (apply mapsto_sub_proper; [now intuition | apply eqtype_subtype]); assumption || now symmetry.
Qed.


Lemma sub_inter_l : forall A B, A ∩ B ⊆ A.
Proof. intros. now left. Qed.

Lemma sub_inter_r : forall A B, A ∩ B ⊆ B.
Proof. intros. now right. Qed.

Lemma sub_inter : forall A B C, C ⊆ A -> C ⊆ B -> C ⊆ A ∩ B.
Proof. intros A B C HA HB ? [? | ?]. now apply HA. now apply HB. Qed.

Lemma inter_sub_proper : forall A A' B B', A ⊆ A' -> B ⊆ B' -> A ∩ B ⊆ A' ∩ B'.
Proof. intros A A' B B' HA HB ? [? | ?]. now left; apply HA. now right; apply HB. Qed.

Lemma inter_mapsto_distr : forall c A B, c ↦ (A ∩ B) ≈ (c ↦ A) ∩ (c ↦ B).
Proof.
intros c A B s. split; intro Hπ.
  destruct Hπ as [Hc [Hπ | Hπ]]. now left; split. now right; split.
  destruct Hπ as [[Hc Hπ] | [Hc Hπ]]. now split; try left. now split; try right.
Qed.

Theorem sub_stack : forall A B π, π ∈ ‖B‖ -> A ⊆ B -> π ∈ ‖A‖.
Proof. auto. Qed.
Arguments sub_stack [A] [B] [π] Hπ HAB.

Theorem sub_term : forall A B t, t ⊩ A -> A ⊆ B -> t ⊩ B.
Proof. intros A B t Ht HAB. start. apply Ht. now apply HAB. Qed.

Theorem sub_id : forall A B, A ⊆ B -> forall e, Id↓e ⊩ A → B.
Proof. intros ? ? H. Ksolve. now apply H. Qed.

(** *** Automatisation and examples **)

(* FIXME: provide good costs to speed up search *)
Create HintDb Ksubtype.
Hint Opaque subtype eqtype : Ksubtype.
Hint Immediate (@reflexivity _ subtype _) (@reflexivity _ eqtype _) sub_Forall sub_bot sub_mapsto : Ksubtype.
Hint Immediate mapsto_cancel mapsto_idempotent inter_mapsto_distr : Ksubtype.
Hint Resolve sub_Impl sub_inter_l sub_inter_r sub_inter sub_Forall_add sub_Forall_cross : Ksubtype.
Hint Resolve Forall_sub_proper mapsto_sub_proper inter_sub_proper : Ksubtype.
Hint Resolve sub_term sub_stack (@transitivity _ subtype _) : Ksubtype.
Hint Extern 1 (_ ⊆ _) => apply top_bot_sub_one; trivial : Ksubtype. (* avoid unfolding of ⊆ *)
Hint Extern 3 (_ ⊆ _) => apply eqtype_subtype : Ksubtype.
Hint Extern 5 (_ ≈ _) => apply subtype_eqtype : Ksubtype.
Hint Extern 7 (_ ≈ _) => apply (@symmetry _ eqtype _) : Ksubtype.   (* high cost to avoid excessive use *)

Ltac subtyping := progress eauto 5 with Ksubtype.

Example inter_idempotent : forall A, A ∩ A ≈ A.
Proof. subtyping. Qed. (* BUG: symmetry used twice for nothing *)

Example callcc_realizes_NNPP : forall e, callcc↓e ⊩ ∀P, (¬¬P) → P.
Proof. intro e. apply (sub_term (callcc_realizes_Peirce e)). subtyping. Qed.

Lemma inter_realizer : forall A B t e, t↓e ⊩ A∩B <-> t↓e ⊩ A /\ t↓e ⊩ B.
Proof.
intros A B t e. split; intro Ht.
  split; subtyping.
  destruct Ht as [Ht₁ Ht₂]. start; (now apply Ht₁) || (now apply Ht₂).
Qed.


(******************************)
(** *  Logical equivalences  **)
(******************************)

(** ** Propositional equivalences **)

(* This lemma is used to have the equivalence between both presentations of equalitiy: = or ≠ primitive *)
Lemma neg_equiv : forall e, (λ"t" λ"c" callcc @ λ"k" "t" @ ("k" @ "c"))↓e ⊩ (⊥ → ⊥) → ∀X, X → X.
Proof. do 2 Ksolve. Qed.

Lemma callcc_other_realizer : forall A B e, callcc↓e ⊩ ((A → B) → ⊥) → A.
Proof. intros A B e. apply (sub_term (callcc_realizes_Peirce e)). subtyping. Qed.

Definition not_impl := λ"t" λ"f" "f" @ (callcc @ "t") @ λ"b" "t" @ λ"a" "b".

Lemma not_impl_realizer : forall e, not_impl↓e ⊩ ∀A B, ¬(A → B) → (A ∧ ¬B).
Proof. Ksolve. startn 1. now eapply callcc_other_realizer; find. do 2 Ksolve. Qed.

Definition not_impl2 :=
  λ"t" λ"f" "f" @ (callcc @ "t") @ (callcc @ λ"g" "t" @ λ"a" "g") @ λ"c" "t" @ λ"a" λ"b" "c".

Lemma not_impl2_realizer : forall e, not_impl2↓e ⊩ ∀A B C, ¬(A → B → C) → (A ∧ B ∧ ¬C).
Proof.
Ksolve.
  startn 1. now eapply callcc_other_realizer; find.
  startn 1. eapply callcc_other_realizer; find. do 2 Ksolve. eok.
  do 2 Ksolve.
Qed.

Definition not_impl3 := λ"t" λ"f" "f" @ (callcc @ "t")
                                      @ (callcc @ (λ"g" "t" @ λ"a" "g"))
                                      @ (callcc @ (λ"g" "t" @ λ"a" λ"b" "g"))
                                      @ λ"d" "t" @ λ"a" λ"b" λ"c" "d".

Lemma not_impl3_realizer : forall e, not_impl3↓e ⊩ ∀A B C D, ¬(A → B → C → D) → (A ∧ B ∧ C ∧ ¬D).
Proof.
Ksolve.
  startn 1. now eapply callcc_other_realizer; find.
  startn 1. eapply (callcc_other_realizer _ (C → D)); find. do 2 Ksolve.
  startn 1. eapply callcc_other_realizer; find. do 2 Ksolve. eok.
  do 2 Ksolve.
Qed.

Definition not_impl4 := λ"t" λ"f" "f" @ (callcc @ "t")
                                      @ (callcc @ (λ"g" "t" @ λ"a" "g"))
                                      @ (callcc @ (λ"g" "t" @ λ"a" λ"b" "g"))
                                      @ (callcc @ (λ"g" "t" @ λ"a" λ"b" λ"c" "g"))
                                      @ λ"e" "t" @ λ"a" λ"b" λ"c" λ"d" "e".

Lemma not_impl4_realizer : forall e, not_impl4↓e ⊩ ∀A B C D E, ¬(A → B → C → D → E) → (A ∧ B ∧ C ∧ D ∧ ¬E).
Proof.
Ksolve.
  startn 1. now eapply callcc_other_realizer; find.
  startn 1. eapply (callcc_other_realizer _ (C → D → E)); find. do 2 Ksolve.
  startn 1. eapply (callcc_other_realizer _ (D → E)); find. do 2 Ksolve.
  startn 1. eapply callcc_other_realizer; find. do 2 Ksolve. eok.
  do 2 Ksolve.
Qed.


Definition De_Morgan_nor := λ"nor" λ"f" "f" @ (λ"a" "nor" @ λ"AZ" λ"BZ" "AZ" @ "a")
                                            @ (λ"b" "nor" @ λ"AZ" λ"BZ" "BZ" @ "b").

Lemma De_Morgan_nor_realizer : forall e, De_Morgan_nor↓e ⊩ ∀A B C, ((A ∨ B) → C) → ((A → C) ∧ (B → C)).
Proof. do 3 Ksolve. Qed.

Corollary De_Morgan_nor_realizer_not : forall e, De_Morgan_nor↓e ⊩ ∀A B, ¬(A ∨ B) → (¬A ∧ ¬B).
Proof. intro e. apply (sub_term (De_Morgan_nor_realizer e)). subtyping. Qed.

Definition De_Morgan_nand :=
  λ"nand" λ"nAZ" λ"nBZ" callcc @ λ"k" "nAZ" @ λ"a" "nand" @ (λ"f" callcc @ λ"k'"
                               ("k" @ ("nBZ" @ λ"b" "k'" @ ("f" @ "a" @ "b")))).

Lemma De_Morgan_nand_realizer : forall e, De_Morgan_nand↓e ⊩ ∀A B C, ((A ∧ B) → C) → ((A → C) ∨ (B → C)).
Proof. do 4 Ksolve. Qed.


(** ** Negation and quantifiers **)

(** ¬∃x F -> ∀x ¬F **)
Theorem not_ex_all_not_proof : forall A (B : A -> Prop), ~(exists x, B x) -> forall x, ~B x.
Proof. intros A B H x Hx. elim H. now exists x. Qed.

Definition not_ex_all_not := λ"Ex" λ"F" "Ex" @ (λ"f" "f" @ "F").

Theorem not_ex_all_not_realizer : forall T (F : T -> formula),
  forall e, not_ex_all_not↓e ⊩ ¬(∃x, F x) → ∀x, ¬F x.
Proof. repeat Ksolve. Qed.

Corollary not_ex_all_not_realizer2 : forall T S (F : T -> S -> formula),
  forall e, not_ex_all_not↓e ⊩ ¬(∃x y, F x y) → ∀x y, ¬F x y.
Proof. repeat Ksolve. Qed.

Corollary not_ex_all_not_realizer3 : forall T S R (F : T -> S -> R -> formula),
  forall e, not_ex_all_not↓e ⊩ ¬(∃x y z, F x y z) → ∀x y z, ¬F x y z.
Proof. repeat Ksolve. Qed.

Corollary not_ex_all_not_realizer4 : forall T S R Q (F : T -> S -> R -> Q -> formula),
  forall e, not_ex_all_not↓e ⊩ ¬(∃x y z t, F x y z t) → ∀x y z t, ¬F x y z t.
Proof. repeat Ksolve. Qed.

(** Relativised version of the preceding theorems **)
Definition not_ex_all_not_rel1 := λ"nEx" λ"x" λ"F" "nEx" @ (λ"f" "f" @ "x" @ "F").

Theorem not_ex_all_not_rel1_realizer : forall T (P F : T -> formula),
  forall e, not_ex_all_not_rel1↓e ⊩ ¬(∃₁x∈P, F x) → ∀₁x∈P, ¬F x.
Proof. repeat Ksolve. Qed.

Definition not_ex_all_not_rel2 :=
  λ"nEx" λ"x" λ"y" λ"F" "nEx" @ (λ"f" "f" @ "x" @ "y" @ "F").

Theorem not_ex_all_not_rel2_realizer :
  forall T₁ T₂ (P₁ : T₁ -> formula) (P₂ : T₂ -> formula) (F : T₁ -> T₂ -> formula),
  forall e, not_ex_all_not_rel2↓e ⊩ ¬(∃₂x,y∈P₁×P₂, F x y) → ∀₂x,y∈P₁×P₂, ¬F x y.
Proof. repeat Ksolve. Qed.

Definition not_ex_all_not_rel3 :=
  λ"nEx" λ"x" λ"y" λ"z" λ"F" "nEx" @ (λ"f" "f" @ "x" @ "y" @ "z" @ "F").

Theorem not_ex_all_not_rel3_realizer :
  forall T₁ T₂ T₃ (P₁ : T₁ -> formula) (P₂ : T₂ -> formula) (P₃ : T₃ -> formula) (F : T₁ -> T₂ -> T₃ -> formula),
  forall e, not_ex_all_not_rel3↓e ⊩ ¬(∃₃x,y,z∈P₁×P₂×P₃, F x y z) → ∀₃x,y,z∈P₁×P₂×P₃, ¬F x y z.
Proof. repeat Ksolve. Qed.

Definition not_ex_all_not_rel4 := λ"nEx" λ"x" λ"y" λ"z" λ"t" λ"F"
  "nEx" @ (λ"f" "f" @ "x" @ "y" @ "z" @ "t" @ "F").

Theorem not_ex_all_not_rel4_realizer : forall T₁ T₂ T₃ T₄ (P₁ : T₁ -> formula) (P₂ : T₂ -> formula)
     (P₃ : T₃ -> formula) (P₄ : T₄ -> formula) (F : T₁ -> T₂ -> T₃ -> T₄ -> formula),
  forall e, not_ex_all_not_rel4↓e ⊩ ¬(∃₄x,y,z,t∈P₁×P₂×P₃×P₄, F x y z t) → ∀₄x,y,z,t∈P₁×P₂×P₃×P₄, ¬F x y z t.
Proof. repeat Ksolve. Qed.


(** ¬∀x F -> ∃x ¬F **)
Theorem not_all_ex_not_proof : forall A (B : A -> Prop), ~(forall x, B x) -> exists x, ~B x.
Proof. intros A B H. apply NNPP. intros Habs. elim H. intro x. apply NNPP. intro. elim Habs. now exists x. Qed.

Definition not_all_ex_not := (* λf. cc (λg. f (cc (λh. g (λi. i h)))) *)
  λ"f" callcc @ (λ"g" "f" @ (callcc @ (λ"h" "g" @ (λ"i" "i" @ "h")))).

Theorem not_all_ex_not_realizer : forall T (F : T -> formula),
  forall e, not_all_ex_not↓e ⊩ ¬(∀x, F x) → ∃x, ¬F x.
Proof. do 3 Ksolve. eok. Qed.

Corollary not_all_ex_not_realizer2 : forall T S (F : T -> S -> formula),
  forall e, not_all_ex_not↓e ⊩ ¬(∀x y, F x y) → ∃x y, ¬F x y.
Proof. do 3 Ksolve. eok. Qed.

Corollary not_all_ex_not_realizer3 : forall T S R (F : T -> S -> R -> formula),
  forall e, not_all_ex_not↓e ⊩ ¬(∀x y z, F x y z) → ∃x y z, ¬F x y z.
Proof. do 3 Ksolve. eok. Qed.

Corollary not_all_ex_not_realizer4 : forall T S R Q (F : T -> S -> R -> Q -> formula),
  forall e, not_all_ex_not↓e ⊩ ¬(∀x y z t, F x y z t) → ∃x y z t, ¬F x y z t.
Proof. do 3 Ksolve. eok. Qed.

(** Relativised version of the preceding theorem **)
Definition not_all_ex_not_rel1 :=
  λ"f" callcc @ (λ"g" ("f" @ (λ"r" callcc @ (λ"h" ("g" @ (λ"i" ("i" @ "r" @ "h"))))))).

Theorem not_all_ex_not_rel1_realizer : forall T (P F : T -> formula),
  forall e, not_all_ex_not_rel1↓e ⊩ ¬(∀₁x∈P, F x) → ∃₁x∈P, ¬F x.
Proof. repeat Ksolve. Qed.

Definition not_all_ex_not_rel2 := λ"f" callcc @ (λ"g" "f" @ (λ"r₁" λ"r₂"
  callcc @ (λ"h" "g" @ (λ"i" "i" @ "r₁" @ "r₂" @ "h")))).

Theorem not_all_ex_not_rel2_realizer :
  forall T₁ T₂ (P₁ : T₁ -> formula) (P₂ : T₂ -> formula) (F : T₁ -> T₂ -> formula),
  forall e, not_all_ex_not_rel2↓e ⊩ ¬(∀₂x,y∈P₁×P₂, F x y) → ∃₂x,y∈P₁×P₂, ¬F x y.
Proof. repeat Ksolve. Qed.

Definition not_all_ex_not_rel3 := λ"f" callcc @ (λ"g" "f" @ (λ"r₁" λ"r₂" λ"r₃"
  callcc @ (λ"h" "g" @ (λ"i" "i" @ "r₁" @ "r₂" @ "r₃" @ "h")))).

Theorem not_all_ex_not_rel3_realizer :
  forall T₁ T₂ T₃ (P₁ : T₁ -> formula) (P₂ : T₂ -> formula) (P₃ : T₃ -> formula) (F : T₁ -> T₂ -> T₃ -> formula),
  forall e, not_all_ex_not_rel3↓e ⊩ ¬(∀₃x,y,z∈P₁×P₂×P₃, F x y z) → ∃₃x,y,z∈P₁×P₂×P₃, ¬F x y z.
Proof. repeat Ksolve. Qed.

Definition not_all_ex_not_rel4 := λ"f" callcc @ (λ"g" "f" @ (λ"r₁" λ"r₂" λ"r₃" λ"r₄"
  callcc @ (λ"h" "g" @ (λ"i" "i" @ "r₁" @ "r₂" @ "r₃" @ "r₄" @ "h")))).

Theorem not_all_ex_not_rel4_realizer : forall T₁ T₂ T₃ T₄ (P₁ : T₁ -> formula) (P₂ : T₂ -> formula)
              (P₃ : T₃ -> formula) (P₄ : T₄ -> formula) (F : T₁ -> T₂ -> T₃ -> T₄ -> formula),
  forall e, not_all_ex_not_rel4↓e ⊩ ¬(∀₄x,y,z,t∈P₁×P₂×P₃×P₄, F x y z t) → ∃₄x,y,z,t∈P₁×P₂×P₃×P₄, ¬F x y z t.
Proof. repeat Ksolve. Qed.

Definition not_all_not_ex := λ"f" λ"g" callcc @ λ"k" "f" @ (λ"a" "k" @ ("g" @ "a")).

Theorem  not_all_not_ex_realizer : forall T (A : T -> formula) e, not_all_not_ex↓e ⊩ (¬(∀x, ¬A x)) → ∃x, A x.
Proof. Ksolve. Ksolve. Qed.


(**************************************************************)
(** *  Prop embedding into stacks: if c then one else ⊤ → ⊥  *)
(**************************************************************)


Set Implicit Arguments.
(* prop (A -> B) ≤ prop A → prop B  and  prop (∀x A(x) ≡ ∀x, prop (A(x)) *)
(* I ⊩ prop (∀x, E₁(x) -> ... -> E_n(x)) ≤ ∀x, prop E₁(x) → ... → prop E_n(x)
   and prop E_i(x) = prop (∀Z, Z(e1) -> Z(e2)) = ∀Z, prop (Z(e1)) → prop (Z(e2)) *)

Definition prop (c : Prop) : formula := fun π => (c /\ π ∈ ‖one‖) \/ (~c /\ π ∈ ‖⊤ → ⊥‖).

Instance prop_proper : Proper (iff ==> eqtype) prop.
Proof. intros A B Heq; repeat split; intros; abstract (compute; rewrite Heq || rewrite <- Heq; assumption). Qed.

Ltac prop_dstack tac Hπ :=
  lazymatch type of Hπ with
    | ?π ∈ ‖prop ?c‖ =>
      unfold prop in Hπ; let HA := fresh "HF" in destruct Hπ as [[HA Hπ] | [HA Hπ]]; try contradiction
    | _ => tac Hπ
  end.
Ltac dstack ::= basic_dstack ltac:(prop_dstack fail).

Lemma distr_subtype : forall A B : Prop, prop (A -> B) ⊆ (prop A → prop B).
Proof.
intros A B. start.
+ left. split. easy. elim (classic A); intro HA.
    exists one. split. start. apply Ht. left. now split. ok.
    exists (⊤ → ⊥). unfold one in *. Ksolve. start. apply Ht. right. repeat split; ok.
+ elim (classic A); intro HA.
    right. split. tauto. split; ok.
    left. split. tauto. exists (⊤ → ⊥). Ksolve. Ksolve. right. repeat split; ok.
Qed.

Lemma Beckman_eqtype : forall A : nat -> Prop, prop (forall n, A n) ≈ (∀n, prop (A n)).
Proof.
intro A. intros s; split; intro Hs; dec s.
  exists 0. left. split. now apply HF. ok.
  apply not_all_ex_not_proof in HF. destruct HF as [n Hn]. exists n. right. repeat split; ok.
  elim (classic (forall n, A n)); intro HA. left. now split. right. unfold one in Hs. dec s. repeat split; ok.
  right. split. intro Habs. elim HF. now apply Habs. split; ok.
Qed.

Lemma stack_prop_iff : forall c t π, t·π ∈ ‖prop c‖ <-> c /\ t★π ∈ ⫫ \/ ~c.
Proof.
intros c t π. split; intro Hπ.
  dec (t·π); try now right. left. split. ok. unfold one in Hπ. dec (t·π). destruct Hπ as [Ht Hπ]. now apply Ht.
  destruct Hπ as [[Hc Hπ] | Hc].
    left. split. ok. exists (fun s => s = π). split. start. compute in Hπ0. now subst. reflexivity.
    right. repeat split; ok.
Qed.

Theorem prop_realizer1 : forall A : Prop, forall e, A -> Id↓e ⊩ prop A.
Proof. intros. start. now apply Id_realizer. Qed.

Theorem prop_realizer2 : forall A : Prop, forall e, ~A -> nId↓e ⊩ ¬(prop A).
Proof. intros A e HA. start. apply Ht. right. repeat split; ok. Qed.

Theorem prop_realizer1' : forall c : Prop, forall t, c -> t ⊩ prop c -> t ⊩ one.
Proof. intros c t Hc Ht. start. apply Ht. left. now split. Qed.

Theorem prop_realizer2' : forall c t, ~c -> t ⊩ prop c -> t ⊩ ⊤ → ⊥.
Proof. intros c t Hc Ht. start. apply Ht. right. repeat split; ok. Qed.

Theorem prop_guard : forall c t t' s, (t ⊩ prop c) -> (c -> t' ★ s ∈ ⫫) -> t ★ t'·s ∈ ⫫.
Proof.
intros c t t' s Ht Ht'. elim (classic c); intro Hc.
  apply (prop_realizer1' Hc) in Ht. apply Ht. exists (fun s' => s' = s). split.
  start. compute in Hπ. subst. now apply Ht'. reflexivity.
  apply (prop_realizer2' Hc) in Ht. apply Ht. split; ok.
Qed.

Theorem guard : forall c F t t' e, (t↓e ⊩ prop c) -> (t'↓e ⊩ F) -> (t @ t')↓e ⊩ (c ↦ F).
Proof. intros c F t t' e Ht Ht'. start. apply Ht. left. split. ok. exists F. split. ok. now apply Hπ. Qed.

Theorem guard_sub : forall c F e π , π ∈ ‖c ↦ F‖ -> Id↓e·π ∈ ‖prop c → F‖.
Proof. intros c F e π Hπ. destruct Hπ as [Hc Hπ]. split; try ok. now apply prop_realizer1. Qed.

Lemma prop_sub_one : forall c, prop c ⊆ one.
Proof. intro c. start. elim (classic c); intro Hc. now left; split. right. split; subtyping. Qed.

Lemma top_bot_sub_prop : forall c, (⊤ → ⊥) ⊆ prop c.
Proof. intro c. start. subtyping. split; ok. Qed.

Theorem prop_subst_stack : forall c c' : Prop, forall s, s ∈ ‖prop c'‖ -> (c -> c') -> s ∈ ‖prop c‖.
Proof.
intros c c' s Hs Hcc'. elim (classic c); intro Hc.
  left. split. ok. dec s. ok. absurd c'; tauto.
  right. split. ok. apply (top_bot_sub_prop Hs).
Qed.

Theorem prop_subtype : forall c c' : Prop, (c -> c') -> prop c ⊆ prop c'.
Proof.
intros c c' Hcc' s [[Hc' Hs] | [Hc' Hs]].
  elim (classic c); intro Hc. left. split; ok. right. unfold one in *. Ksolve.
  right. split. now auto. ok.
Qed.

Hint Immediate prop_sub_one top_bot_sub_prop : Ksubtype.
Hint Resolve distr_subtype prop_subtype Beckman_eqtype : Ksubtype.

Theorem prop_subst_term : forall c c' : Prop, forall t, t ⊩ prop c -> (c -> c') -> t ⊩ prop c'.
Proof. subtyping. Qed.


Lemma prop_term_destruct : forall t F, t ⊩ prop F -> F /\ t ⊩ one \/ ~F /\ t ⊩ ⊤ → ⊥.
Proof.
intros t F Ht. elim (classic F); intro.
  left. subtyping.
  right. split. ok. eapply prop_realizer2'; eassumption.
Qed.

Ltac destruct_prop Ht :=
  match type of Ht with
    | ?t ⊩ prop ?a =>
      apply prop_term_destruct in Ht; let Hc := fresh "Hc" in destruct Ht as [[Hc Ht] | [Hc Ht]]; try contradiction
  end.

Example chain2_prop : forall c₁ c₂ : Prop, forall e, (c₁ -> c₂) -> Id↓e ⊩ prop c₁ → prop c₂.
Proof. intros c₁ c₂ e Hcs. apply (sub_term (prop_realizer1 _ Hcs)). subtyping. Qed.

Example chain3_prop : forall c₁ c₂ c₃ : Prop, forall e, (c₁ -> c₂ -> c₃) -> Id↓e ⊩ prop c₁ → prop c₂ → prop c₃.
Proof. intros c₁ c₂ c₃ e Hcs. apply (sub_term (prop_realizer1 _ Hcs)). subtyping. Qed.

Example chain4_prop :
  forall c₁ c₂ c₃ c₄ : Prop, forall e, (c₁ -> c₂ -> c₃ -> c₄) -> Id↓e ⊩ prop c₁ → prop c₂ → prop c₃ → prop c₄.
Proof. intros c₁ c₂ c₃ c₄ e Hcs. apply (sub_term (prop_realizer1 _ Hcs)). subtyping. Qed.

(* The converse is false : ~(c₁ -> c₂) because t ⊩ c₁ → c₂ does not implies t ⊩ ⊤ → ⊥
   (⫫  is not closed under red) *)
Global Opaque prop.

Definition BB1 := λ"t" λ"u" callcc @ λ"k" "t" @ ("k" @ "u").

Theorem BB1_realizer : forall e, BB1↓e ⊩ ((⊥ → ⊥) → one) ∩ ((⊤ → ⊥) → ⊤ → ⊥).
Proof. unfold one. do 2 Ksolve. Qed.
